#include "IOHandler.h"
#include "StringTokenizer.h"

void trim(std::string& s)  {
  if ( s.empty() )
    return;
  std::string::iterator p;

  for(p=s.end(),p--; p!= s.begin() && (*p == ' ' || *p == '\t');p--)
    ;
  if ( *p != ' ' && *p != '\t' ) 
    p++;
  s.erase(p,s.end());

  for(p=s.begin(); p!= s.end() && (*p == ' ' || *p == '\t');p++)
    ;
  s.erase(s.begin(),p);

}

int IOHandler::readParams(char *filename)  {
  ostringstream errstream;
  errstream << "[IOHandler::readParams]  ";

  ifstream parin;
  parin.open(filename,ios::in);
  if ( !parin.is_open() )  {
    errstream << "can't open file " << filename;
    throw(Exception(errstream.str()));
  }

  // universal defaults
  setParam("qpheno","NO");
  setParam("hsimple","NO");
  setParam("oddsratio","NO");
  setParam("effectsize","NO");
  setParam("useempiricalvariance","NO");
  setParam("genoformat","integer");

  // setParam("dosages","NO");
  setParam("adjustment","NO");

  while ( !parin.eof() )  {
    string line, origline, pname, pval;

    getline(parin,line);
    origline = line;

    if ( line.empty() && parin.eof() )  
      break;

    StringTokenizer st(line,":");
    if ( (int) st.countTokens() != 2 )  {
      errstream << "Can't parse parameter line " << origline;
      throw(Exception(errstream.str()));
    }

    st.nextToken(pname);
    st.nextToken(pval);

    // remove leading and trailing white space 
    trim(pname);
    trim(pval);
    
    if ( !pname.empty() && !pval.empty() )  {
      setParam(pname,pval);
    }
    else  {
      errstream << "Can't parse parameter line " << origline;
      throw(Exception(errstream.str()));
    }

  }
  parin.close();

  string str;
  if ( !getParam("outfile",&str) )  {
    setParam("output2file","YES");
  }
  else  {
    setParam("output2file","NO");
  }

  if ( !getParam("ancname",&str) )  {
    setParam("ancname0",str+".0");
    setParam("ancname1",str+".1");
    setParam("ancname2",str+".2");
  }

  if ( !getParam("orfile",&str) )  {
    setParam("oddsratio","YES");
  }

  if ( !getParam("esfile",&str) )  {
    setParam("effectsize","YES");
  }

  return 0;
}

bool IOHandler::previewParams(char *filename)  {
  ostringstream errstream;
  errstream << "[IOHandler::readParams]  ";

  ifstream parin;
  parin.open(filename,ios::in);
  if ( !parin.is_open() )  {
    errstream << "can't open file " << filename;
    throw(Exception(errstream.str()));
  }

  while ( !parin.eof() )  {
    string line, origline, pname, pval;

    getline(parin,line);
    origline = line;

    if ( line.empty() && parin.eof() )  
      break;

    StringTokenizer st(line,":");
    if ( (int) st.countTokens() != 2 )  {
      errstream << "Can't parse parameter line " << origline;
      throw(Exception(errstream.str()));
    }

    st.nextToken(pname);
    st.nextToken(pval);

    // remove leading and trailing white space 
    trim(pname);
    trim(pval);
    
    if ( !pname.empty() && !pval.empty() )  {
      if ( !pname.compare("useempiricalvariance") )  {
         if ( !pval.compare("YES")  || !pval.compare("yes") ||
              !pval.compare("TRUE") || !pval.compare("true") ||
              !pval.compare("1") )  {

              parin.close();
              return true;
          }
      }
    }
    else  {
      errstream << "Can't parse parameter line " << origline;
      throw(Exception(errstream.str()));
    }

  }

  parin.close();
  return false;

}



int IOHandler::readIntLine(ifstream *f, int ndata, int **data)  {
  ostringstream errstream;
  errstream << "[IOHandler::readIntLine]  ";

  if ( !(*f).is_open() )  { 
    errstream << "file is not open";
    throw(Exception(errstream.str()));
  }

  string line;
  getline(*f,line);

  if ( (*f).eof() ) {
    errstream << "premature end-of-file";
    throw(Exception(errstream.str()));
  }

  if ( (int) line.length() < ndata )  {
    errstream << "line too short";
    throw(Exception(errstream.str()));
  }

  if ( (int) line.length() > ndata )  {
    cerr << "(warning) line too long" << endl;
  }


  for(int i=0;i<ndata;i++)  {
    int k = (int) line[i];
    k -= 48;                     // 48 is ASCII('0')
    if ( 0 <= k && k <= 9 )  {
      (*data)[i] = k;
    }
    else  {
      (*data)[i] = 9;
    }
  }
  return 0; 

}


int IOHandler::readOneColumnFile(string filename, int ndata, double **data)  {
  ostringstream errstream;
  errstream << "[IOHandler::readRealLine]  ";

  ifstream instream;
  instream.open(filename.c_str(),ios::in);
  if ( !instream.is_open() )  {
    errstream << "can't open globalanc file" << filename;
    throw(Exception(errstream.str()));
  }

  for(int i=0;i<ndata;i++)  {
 
    string line;
    getline(instream,line);

    if ( instream.eof() ) {
      errstream << "premature end of file";
      throw(Exception(errstream.str()));
    }

    StringTokenizer st(line," \t");

    if ( (int) st.countTokens() > 1 )  {
      cerr << "(warning) too many tokens on input line\n";
    }

    if ( (int) st.countTokens() < 1 )  {
      errstream << "too few tokens on input line";
      throw(Exception(errstream.str()));
    }

    string s;
    st.nextToken(s);
    stringstream ss(s);
    double x;
    ss >> x;
    if ( ss.fail() )  {
      errstream << "incomprehensible input";
      throw(Exception(errstream.str()));
    }
    (*data)[i] = x;
  }

  instream.close();
  return 0;
}

int IOHandler::readRealLine(ifstream *f, int ndata, double **data)  {

  ostringstream errstream;
  errstream << "[IOHandler::readRealLine]  ";


  if ( !(*f).is_open() )  { 
    errstream << "file is not open";
    throw(Exception(errstream.str()));
  }
 
  string line;
  getline(*f,line);

  if ( (*f).eof() ) {
    errstream << "premature end of file";
    throw(Exception(errstream.str()));
  }

  StringTokenizer st(line," \t");

  if ( (int) st.countTokens() > ndata )  {
    cerr << "(warning) too many tokens on input line\n";
  }

  if ( (int) st.countTokens() < ndata )  {
    errstream << "too few tokens on input line";
    throw(Exception(errstream.str()));
  }

  for(int i=0;i<ndata;i++)  {
    string s;
    st.nextToken(s);
    stringstream ss(s);
    double x;
    ss >> x;
    if ( ss.fail() )  {
      errstream << "incomprehensible input";
      throw(Exception(errstream.str()));
    }
    (*data)[i] = x;
  }

  return 0;
}

int IOHandler::getParam(string key, string *valp)  {
  map<string,string>::iterator p;
  p = paramMap.find(key);
  if ( p != paramMap.end() )  {
    *valp = (*p).second;
    return 0;
  }
  else  {
    return 1;
  }
}

int IOHandler::getPositiveIntParam(string key)  {
  ostringstream errstream;
  errstream << "[IOHandler::getPositiveIntParam] ";
  string val;
  if ( !getParam(key,&val) )  {
    int n = atoi(val.c_str());
    if ( 0 < n )  {
      return n;
    }
    else  {
      errstream << "parameter " << key << " not a positive integer";
      throw(Exception(errstream.str()));
    }
  }
  else  {
    errstream << "parameter " << key << " not found";
    throw(Exception(errstream.str()));
  }
}

string IOHandler::getStringParam(string key)  {
  ostringstream errstream;
  errstream << "[IOHandler::getStringParam] ";
  string val;
  if ( getParam(key,&val) )  {
    errstream << "parameter " << key << " not found";
    throw(Exception(errstream.str()));
  }
  return val;
}

bool IOHandler::getBoolParam(string key)  {
  ostringstream errstream;
  errstream <<  "[IOHandler::getBoolParam] ";
  string val;
  if ( getParam(key,&val) )  {
    errstream << "parameter " << key << " not found";
    throw(Exception(errstream.str()));
  }

  if ( !val.compare("YES") || !val.compare("yes") ||
       !val.compare("TRUE") || !val.compare("true") ||
       !val.compare("1") )  {
    return true;
  }
  else  {
    return false;
  }
}

void IOHandler::setParam(string key, string val)  {
  map<string,string>::iterator p;
  p = paramMap.find(key);
  if ( p != paramMap.end() )  {  // change
    (*p).second = val;
  }
  else  {
    paramMap[key] = val;
  }
}

void IOHandler::dbgShowParams(void)  {
  map<string, string>::iterator p;
  for(p=paramMap.begin();p != paramMap.end(); p++)  {
    cout << ":" << p->first << ":" << p->second << ":" << endl;
  }
}

void IOHandler::openInput(string param, ifstream *s)  {
  ostringstream errstream;
  errstream << "[IOHandler::openInput] ";

  string filename;
  getParam(param, &filename);

  (*s).open(filename.c_str(),ios::in);
  if ( !(*s).is_open() )  {
    errstream << "can't open file " << filename << " for key " << param;
    throw(Exception(errstream.str()));
  }
  return;
}

void IOHandler::openOutput(string param, ofstream *s)  {
  ostringstream errstream;
  errstream << "[IOHandler::openOutput] ";

  string filename;
  getParam(param, &filename);
  (*s).open(filename.c_str(),ios::out);
  if ( !(*s).is_open() )  {
    errstream << "can't open file " << filename;
    throw(Exception(errstream.str()));
  }
  return;
}
