#ifndef _SCORESAS_
#define _SCORESAS_

#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <cmath>

#include <gsl/gsl_errno.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_min.h>

#include "IOHandler.h"
#include "Exception.h"

using namespace std;

struct sasparams  {
  double XAARRT, XAARVT, XAAVVT, XAARRS, XAARVS, XAAVVS;
  double XAERRT, XAERVT, XAEVVT, XAERRS, XAERVS, XAEVVS;
  double XEERRT, XEERVT, XEEVVT, XEERRS, XEERVS, XEEVVS;
  double XART,XAVT,XARS,XAVS;
  double XERT,XEVT,XERS,XEVS;
  double R, pE, pA;
  double rhatA, rhatE;
  double impPa,impPe;
  double *TS, *AS0, *AS1, *AS2;
  int *AS;
  int nt;
  bool local_anc_is_prob;
};
typedef struct sasparams sasparams_t;

class ScoreSAS  {

  public:

    // constructor
    ScoreSAS(char *paramfile);

    // destructor
    ~ScoreSAS(void);

    // main loop
    int evaluate(void);

  private:

    int nsamples;
    int nsnps;
    int errflag;  // remove when exception handling introduced

    bool geno_is_dosage;
    bool use_adjustment;

    ifstream genoStream;
    ifstream phenoStream;
    ifstream localAncStream;
    ofstream outputStream;
    ofstream oddsratioStream;

    int    *dipGenoVec;
    int    *dipPhenoVec;
    int    *dipLocalAnc;

    string genoform;
    double* genoDosages;
    double *globalAnc;

    IOHandler ioh;

    string oddsratio;
    void setOddsRatio(double r);

    sasparams_t *mysasparams;

    double evalSAS();

    // overloaded parameter evaluators 
    void init_sasparams(double *D, int *P, int *A, double *T);
    void init_sasparams(int *G, int *P, int *A, double *T);

    // print parameters for exception message
    string errorString(void);

    // backup for when Brent's algorithm fails
    int binsearch(double(*f)(double, sasparams_t *), 
      sasparams_t *p, double xmin, double xmax, double *xout);

    double getSASEstimatedImputationR2(double *dosage, int *P, int *A, int nsamples, int anc);
    double getSASEstimatedImputationR2(int *g, int *P, int *A, int nsamples, int anc);

};

#endif
