#ifndef _SCORESASGC_
#define _SCORESASGC_

#include <iostream>
#include <iomanip>
#include <fstream>
#include <string>
#include <cmath>

#include <gsl/gsl_errno.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_min.h>

#include "IOHandler.h"
#include "Exception.h"

using namespace std;

struct sasparams_gc  {
  double XAARRT, XAARVT, XAAVVT, XAARRS, XAARVS, XAAVVS;
  double XAERRT, XAERVT, XAEVVT, XAERRS, XAERVS, XAEVVS;
  double XEERRT, XEERVT, XEEVVT, XEERRS, XEERVS, XEEVVS;
  double XART,XAVT,XARS,XAVS;
  double XERT,XEVT,XERS,XEVS;
  double R, pE, pA;
  double rhatA, rhatE;
  double impPa,impPe;
  double *TS, *AS0, *AS1, *AS2;
  int *AS;
  int nt;
  double sdGamma;
  double meanGamma;
};
typedef struct sasparams_gc sasparams_gc_t;

class ScoreSASGC  {

  public:

    // constructor
    ScoreSASGC(char *paramfile);

    // destructor
    ~ScoreSASGC(void);

    // main loop
    int evaluate(void);
    void computeMeanSdGamma(int *p);

  private:

    int nsamples;
    int nsnps;
    int errflag;  // remove when exception handling introduced

    bool geno_is_dosage;
    bool use_adjustment;

    ifstream genoStream;
    ifstream phenoStream;
    ifstream localAncStream;
    ofstream outputStream;
    ofstream oddsratioStream;

    int    *dipGenoVec;
    int    *dipPhenoVec;
    int    *dipLocalAnc;

    double* genoDosages;
    double *globalAnc;

    double sdGamma;
    double meanGamma;

    IOHandler ioh;

    string oddsratio;
    void setOddsRatio(double r);

    sasparams_gc_t *mysasparams_gc;

    double evalSAS();

    // overloaded parameter evaluators 
    void init_sasparams_gc(double *D, int *P, int *A, double *T);
    void init_sasparams_gc(int *G, int *P, int *A, double *T);

    // print parameters for exception message
    string errorString(void);
    double loglik(sasparams_gc_t *p, double pA, double pE, double R );
    // backup for when Brent's algorithm fails
    int binsearch(double(*f)(double, sasparams_gc_t *), 
      sasparams_gc_t *p, double xmin, double xmax, double *xout);

    // auxiliary functions for solving maximization problem
    double getSASEstimatedImputationR2(double *dosage, int *P, int *A, int nsamples, int anc);
    double getSASEstimatedImputationR2(int *g, int *P, int *A, int nsamples, int anc);


    // double C( sasparams_gc_t *sasparams_gc );
    // double dCdR( sasparams_gc_t *sasparams_gc );
    // double dCdpA( sasparams_gc_t *sasparams_gc );
    // double dCdpE( sasparams_gc_t *sasparams_gc );

};

#endif
